import os
from collections.abc import AsyncIterator, Iterator

import pytest
from sqlalchemy import text
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from testcontainers.postgres import PostgresContainer

from tests.v1.test_integration.mocks.model import MockModel
from tests.v1.test_integration.mocks.repo import MockRepo
from tests.v1.test_integration.mocks.schema import MockModelResponseSchema
from tests.v1.test_integration.mocks.service import MockService

DB_DSN = os.getenv(
    'DB_DSN',
    'postgresql+asyncpg://notora:notora@tests_db:5432/notora',
)


@pytest.fixture(scope='session')
def postgres_db() -> Iterator[PostgresContainer]:
    with PostgresContainer() as db:
        yield db


@pytest.fixture(scope='session')
async def db_engine(postgres_db: PostgresContainer) -> AsyncIterator[AsyncEngine]:
    url = postgres_db.get_connection_url(driver='asyncpg')
    engine = create_async_engine(url)
    yield engine
    await engine.dispose()


@pytest.fixture(scope='session', autouse=True)
async def init_db(db_engine: AsyncEngine) -> None:
    """Restart all sequences from database.

    Called one time for all tests.
    Cleans up "global_id_sequence" in addition to table sequences,
    so just TRUNCATE TABLE ... RESTART IDENTITY CASCADE wouldn't be enough.
    """
    async with db_engine.begin() as conn:
        await conn.run_sync(MockModel.metadata.create_all)
        stmt = text("SELECT c.relname FROM pg_class c WHERE c.relkind = 'S';")
        sequences = (await conn.execute(stmt)).scalars().all()
        for sequence in sequences:
            await conn.execute(text(f'ALTER SEQUENCE {sequence} RESTART;'))


@pytest.fixture(autouse=True)
async def clean_all_tables(db_engine: AsyncEngine) -> None:
    """Clean all tables before tests and after every test."""
    stmt = text("SELECT t.table_name FROM information_schema.tables t WHERE table_schema='public'")
    async with db_engine.begin() as conn:
        tables = (await conn.execute(stmt)).scalars().all()
        tables = [
            t_name
            for t_name in tables
            if not t_name.startswith('pg_')
            and not t_name.startswith('v_')
            and not t_name.startswith('alembic')
        ]
        for table_name in tables:
            await conn.execute(text(f'TRUNCATE TABLE "{table_name}" CASCADE;'))


@pytest.fixture(scope='session')
def session_factory(db_engine: AsyncEngine) -> async_sessionmaker[AsyncSession]:
    return async_sessionmaker(db_engine, expire_on_commit=False)


@pytest.fixture(scope='function')
async def db_session(
    session_factory: async_sessionmaker[AsyncSession],
    db_engine: AsyncEngine,
) -> AsyncIterator[AsyncSession]:
    async with session_factory() as session:
        try:
            yield session
        finally:
            await session.rollback()


@pytest.fixture(scope='function')
def mock_repo() -> MockRepo:
    return MockRepo(MockModel)


@pytest.fixture(scope='function')
def mock_service(mock_repo: MockRepo) -> MockService:
    service = MockService(mock_repo)
    service.response_schema = MockModelResponseSchema
    return service



from uuid import UUID

from notora.v2.schemas.base import BaseRequestSchema, BaseResponseSchema


class V2UserCreateSchema(BaseRequestSchema):
    id: UUID
    email: str
    name: str
    is_active: bool = True


class V2UserResponseSchema(BaseResponseSchema):
    id: UUID
    email: str
    name: str
    is_active: bool
    updated_by: UUID | None = None


class V2ProfileResponseSchema(BaseResponseSchema):
    id: UUID
    user_id: UUID
    bio: str



from collections.abc import Iterable
from typing import Any, overload

from sqlalchemy import Executable
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.params import PaginationParams
from notora.v2.repositories.types import FilterSpec, OptionSpec, OrderSpec
from notora.v2.schemas.base import (
    BaseResponseSchema,
    PaginatedResponseSchema,
)
from notora.v2.services.mixins.listing import ListingServiceMixin

__all__ = ['PaginationServiceMixin']

class PaginationServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    ListingServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
):
    __type_params__: tuple[object, ...]

    @overload
    async def paginate(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int = 20,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
        schema: None = ...,
    ) -> PaginatedResponseSchema[ListSchema]: ...
    @overload
    async def paginate[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int = 20,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
        schema: type[SchemaT],
    ) -> PaginatedResponseSchema[SchemaT]: ...
    @overload
    async def paginate_from_queries(
        self,
        session: AsyncSession,
        *,
        data_query: Executable,
        count_query: Executable,
        limit: int,
        offset: int,
        schema: None = ...,
    ) -> PaginatedResponseSchema[ListSchema]: ...
    @overload
    async def paginate_from_queries[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        *,
        data_query: Executable,
        count_query: Executable,
        limit: int,
        offset: int,
        schema: type[SchemaT],
    ) -> PaginatedResponseSchema[SchemaT]: ...
    @overload
    async def paginate_params(
        self,
        session: AsyncSession,
        params: PaginationParams[ModelType],
        *,
        schema: None = ...,
    ) -> PaginatedResponseSchema[ListSchema]: ...
    @overload
    async def paginate_params[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        params: PaginationParams[ModelType],
        *,
        schema: type[SchemaT],
    ) -> PaginatedResponseSchema[SchemaT]: ...



import pytest

from notora.v2.schemas.base import PaginationMetaSchema


@pytest.mark.parametrize('limit', [0, -1])
def test_pagination_meta_rejects_non_positive_limit(limit: int) -> None:
    with pytest.raises(ValueError, match='limit must be a positive integer'):
        PaginationMetaSchema.calculate(total=10, limit=limit, offset=0)


@pytest.mark.parametrize('offset', [-1, -10])
def test_pagination_meta_rejects_negative_offset(offset: int) -> None:
    with pytest.raises(ValueError, match='offset must be zero or a positive integer'):
        PaginationMetaSchema.calculate(total=10, limit=5, offset=offset)


def test_pagination_meta_zero_total_is_single_page() -> None:
    total = 0
    limit = 10
    offset = 0
    meta = PaginationMetaSchema.calculate(total=total, limit=limit, offset=offset)
    assert meta.total == total
    assert meta.limit == limit
    assert meta.offset == offset


def test_pagination_meta_retains_offset_and_total() -> None:
    total = 5
    limit = 2
    offset = 10
    meta = PaginationMetaSchema.calculate(total=total, limit=limit, offset=offset)
    assert meta.total == total
    assert meta.limit == limit
    assert meta.offset == offset



from collections.abc import Iterable
from typing import Protocol, overload

from notora.v2.models.base import GenericBaseModel
from notora.v2.schemas.base import BaseResponseSchema

__all__ = ['SerializerMixin', 'SerializerProtocol']

class SerializerProtocol[
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](Protocol):
    __type_params__: tuple[object, ...]

    @overload
    def serialize_one(
        self,
        obj: ModelType,
        *,
        schema: None = ...,
    ) -> DetailSchema: ...
    @overload
    def serialize_one[SchemaT: BaseResponseSchema](
        self,
        obj: ModelType,
        *,
        schema: type[SchemaT],
    ) -> SchemaT: ...
    @overload
    def serialize_many(
        self,
        objs: Iterable[ModelType],
        *,
        schema: None = ...,
        prefer_list_schema: bool = True,
    ) -> list[ListSchema]: ...
    @overload
    def serialize_many[SchemaT: BaseResponseSchema](
        self,
        objs: Iterable[ModelType],
        *,
        schema: type[SchemaT],
        prefer_list_schema: bool = True,
    ) -> list[SchemaT]: ...

class SerializerMixin[
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
]:
    __type_params__: tuple[object, ...]
    detail_schema: type[DetailSchema] | None
    list_schema: type[ListSchema] | None

    @overload
    def serialize_one(
        self,
        obj: ModelType,
        *,
        schema: None = ...,
    ) -> DetailSchema: ...
    @overload
    def serialize_one[SchemaT: BaseResponseSchema](
        self,
        obj: ModelType,
        *,
        schema: type[SchemaT],
    ) -> SchemaT: ...
    @overload
    def serialize_many(
        self,
        objs: Iterable[ModelType],
        *,
        schema: None = ...,
        prefer_list_schema: bool = True,
    ) -> list[ListSchema]: ...
    @overload
    def serialize_many[SchemaT: BaseResponseSchema](
        self,
        objs: Iterable[ModelType],
        *,
        schema: type[SchemaT],
        prefer_list_schema: bool = True,
    ) -> list[SchemaT]: ...



"""Repository mixins and base classes for v2."""

from notora.v2.repositories.base import (
    Repository,
    RepositoryProtocol,
    SoftDeleteRepository,
    SoftDeleteRepositoryProtocol,
)
from notora.v2.repositories.config import RepoConfig
from notora.v2.repositories.factory import build_repository
from notora.v2.repositories.mixins import (
    CountableMixin,
    CreateMixin,
    CreateOrSkipMixin,
    DeleteMixin,
    FilterableMixin,
    ListableMixin,
    LoadOptionsMixin,
    OrderableMixin,
    PrimaryKeyMixin,
    RetrievableMixin,
    SelectableMixin,
    SoftDeleteMixin,
    UpdateMixin,
    UpsertMixin,
)
from notora.v2.repositories.params import PaginationParams, QueryParams
from notora.v2.repositories.query_dsl import (
    FilterField,
    FilterOperator,
    QueryInput,
    SortDirection,
    SortField,
    build_query_params,
    make_query_params_dependency,
)
from notora.v2.repositories.types import (
    DEFAULT_LIMIT,
    DefaultLimit,
    FilterClause,
    FilterFactory,
    FilterSpec,
    OptionFactory,
    OptionSpec,
    OrderClause,
    OrderFactory,
    OrderSpec,
    SupportsOptions,
    SupportsWhere,
)

__all__ = [
    'DEFAULT_LIMIT',
    'CountableMixin',
    'CreateMixin',
    'CreateOrSkipMixin',
    'DefaultLimit',
    'DeleteMixin',
    'FilterClause',
    'FilterFactory',
    'FilterField',
    'FilterOperator',
    'FilterSpec',
    'FilterableMixin',
    'ListableMixin',
    'LoadOptionsMixin',
    'OptionFactory',
    'OptionSpec',
    'OrderClause',
    'OrderFactory',
    'OrderSpec',
    'OrderableMixin',
    'PaginationParams',
    'PrimaryKeyMixin',
    'QueryInput',
    'QueryParams',
    'RepoConfig',
    'Repository',
    'RepositoryProtocol',
    'RetrievableMixin',
    'SelectableMixin',
    'SoftDeleteMixin',
    'SoftDeleteRepository',
    'SoftDeleteRepositoryProtocol',
    'SortDirection',
    'SortField',
    'SupportsOptions',
    'SupportsWhere',
    'UpdateMixin',
    'UpsertMixin',
    'build_query_params',
    'build_repository',
    'make_query_params_dependency',
]



from math import ceil
from uuid import UUID

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v1.schemas.base import Filter
from tests.v1.test_integration.mocks.model import MockModel, MockModelFactory
from tests.v1.test_integration.mocks.schema import (
    CreateMockModelRequestSchema,
    MockModelResponseSchema,
    UpdateMockModelRequestSchema,
)
from tests.v1.test_integration.mocks.service import MockService


async def _setup(db_session: AsyncSession, count: int) -> list[MockModel]:
    objs = [MockModelFactory.build() for _ in range(count)]
    db_session.add_all(objs)
    await db_session.commit()
    return objs


async def _select_one(db_session: AsyncSession, pk: UUID) -> MockModel:
    query = select(MockModel).where(MockModel.id == pk)
    return (await db_session.scalars(query)).one()


def test_serialize(mock_service: MockService) -> None:
    db_obj = MockModelFactory.build()

    schema_obj = mock_service.serialize(db_obj)

    assert isinstance(schema_obj, MockModelResponseSchema)


def test_serialize_many(mock_service: MockService) -> None:
    objs_count = 5
    db_objs = [MockModelFactory.build() for _ in range(objs_count)]

    schema_objs = mock_service.serialize_many(db_objs)

    assert len(schema_objs) == objs_count
    assert all(isinstance(obj, MockModelResponseSchema) for obj in schema_objs)


async def test_list(db_session: AsyncSession, mock_service: MockService) -> None:
    objs_count = 5
    await _setup(db_session, objs_count)

    items = await mock_service.list_(db_session, limit=objs_count)

    assert len(items) == objs_count
    assert all(isinstance(item, MockModelResponseSchema) for item in items)


async def test_paginate(db_session: AsyncSession, mock_service: MockService) -> None:
    limit = 5
    offset = 0
    objs_count = 10
    await _setup(db_session, objs_count)

    res = await mock_service.paginate(db_session, limit=limit, offset=offset)

    assert res.meta.limit == limit
    assert res.meta.total == objs_count
    assert res.meta.current_page == 1
    assert res.meta.last_page == ceil(objs_count / limit)
    assert len(res.data) == limit
    assert all(isinstance(item, MockModelResponseSchema) for item in res.data)


async def test_create(db_session: AsyncSession, mock_service: MockService) -> None:
    new_data = CreateMockModelRequestSchema(name='test')

    created_obj = await mock_service.create(db_session, new_data)

    res = await _select_one(db_session, created_obj.id)
    assert res.id is not None


async def test_retrieve(db_session: AsyncSession, mock_service: MockService) -> None:
    objs = await _setup(db_session, 3)
    target_obj = objs[0]

    retrieved_obj = await mock_service.retrieve(db_session, target_obj.id)

    res = await _select_one(db_session, retrieved_obj.id)
    assert res.id == retrieved_obj.id


async def test_retrieve_one_by(db_session: AsyncSession, mock_service: MockService) -> None:
    objs = await _setup(db_session, 3)
    target_obj = objs[0]

    retrieved_obj = await mock_service.retrieve_one_by(
        db_session, filters=[Filter(field='id', op='=', value=target_obj.id)]
    )

    res = await _select_one(db_session, retrieved_obj.id)
    assert res.id == retrieved_obj.id


async def test_update(db_session: AsyncSession, mock_service: MockService) -> None:
    objs = await _setup(db_session, 1)
    target_obj = objs[0]
    new_data = UpdateMockModelRequestSchema(name='test')

    await mock_service.update(db_session, new_data, target_obj.id)

    res = await _select_one(db_session, target_obj.id)
    assert res.name == new_data.name


async def test_soft_delete(db_session: AsyncSession, mock_service: MockService) -> None:
    objs = await _setup(db_session, 1)
    target_obj = objs[0]

    deleted_obj = await mock_service.soft_delete(db_session, target_obj.id)

    assert deleted_obj is not None
    assert isinstance(deleted_obj, MockModel)
    assert deleted_obj.deleted_at is not None


async def test_soft_delete_one_by(db_session: AsyncSession, mock_service: MockService) -> None:
    objs = await _setup(db_session, 1)
    target_obj = objs[0]

    deleted_obj = await mock_service.soft_delete_one_by(
        db_session, filters=[Filter(field='id', op='=', value=target_obj.id)]
    )

    assert deleted_obj is not None
    assert isinstance(deleted_obj, MockModel)
    assert deleted_obj.deleted_at is not None



from uuid import uuid4

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.exceptions.common import AlreadyExistsError, FKNotFoundError
from tests.v2.test_integration.mocks.service import V2ProfileService, V2UserService


async def test_unique_violation_translates(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    payload = {
        'id': uuid4(),
        'email': 'unique@ex.com',
        'name': 'Unique',
        'is_active': True,
    }
    await user_service.create(db_session, payload)
    await db_session.commit()

    duplicate_payload = {
        'id': uuid4(),
        'email': 'unique@ex.com',
        'name': 'Duplicate',
        'is_active': True,
    }
    with pytest.raises(AlreadyExistsError):
        await user_service.create(db_session, duplicate_payload)


async def test_fk_violation_translates(
    db_session: AsyncSession,
    profile_service: V2ProfileService,
) -> None:
    payload = {
        'id': uuid4(),
        'user_id': uuid4(),
        'bio': 'profile',
    }

    with pytest.raises(FKNotFoundError) as exc:
        await profile_service.create(db_session, payload)

    assert exc.value.table_name == 'v2_profile'



from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.create import CreateOrSkipServiceMixin, CreateServiceMixin
from notora.v2.services.mixins.delete import DeleteServiceMixin, SoftDeleteServiceMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.listing import ListingServiceMixin
from notora.v2.services.mixins.m2m import M2MSyncMode, ManyToManyRelation, ManyToManySyncMixin
from notora.v2.services.mixins.pagination import PaginationServiceMixin
from notora.v2.services.mixins.payload import PayloadMixin
from notora.v2.services.mixins.retrieval import RetrievalServiceMixin
from notora.v2.services.mixins.serializer import SerializerMixin, SerializerProtocol
from notora.v2.services.mixins.update import UpdateByFilterServiceMixin, UpdateServiceMixin
from notora.v2.services.mixins.updated_by import UpdatedByServiceMixin
from notora.v2.services.mixins.upsert import UpsertServiceMixin

__all__ = [
    'CreateOrSkipServiceMixin',
    'CreateServiceMixin',
    'DeleteServiceMixin',
    'ListingServiceMixin',
    'M2MSyncMode',
    'ManyToManyRelation',
    'ManyToManySyncMixin',
    'PaginationServiceMixin',
    'PayloadMixin',
    'RepositoryAccessorMixin',
    'RetrievalServiceMixin',
    'SerializerMixin',
    'SerializerProtocol',
    'SessionExecutorMixin',
    'SoftDeleteServiceMixin',
    'UpdateByFilterServiceMixin',
    'UpdateServiceMixin',
    'UpdatedByServiceMixin',
    'UpsertServiceMixin',
]



from collections.abc import Callable
from typing import Any, Protocol, Self

from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.base import ExecutableOption
from sqlalchemy.sql.expression import UnaryExpression

from notora.v2.models.base import GenericBaseModel


class _DefaultLimit:
    def __repr__(self) -> str:
        return 'DEFAULT_LIMIT'


DEFAULT_LIMIT = _DefaultLimit()
type DefaultLimit = _DefaultLimit

type FilterClause = ColumnElement[bool]
type FilterFactory[ModelType: GenericBaseModel] = Callable[[type[ModelType]], FilterClause]
type FilterSpec[ModelType: GenericBaseModel] = FilterClause | FilterFactory[ModelType]

type OrderClause = ColumnElement[Any] | UnaryExpression[Any]
type OrderFactory[ModelType: GenericBaseModel] = Callable[[type[ModelType]], OrderClause]
type OrderSpec[ModelType: GenericBaseModel] = OrderClause | OrderFactory[ModelType]

type OptionFactory[ModelType: GenericBaseModel] = Callable[[type[ModelType]], ExecutableOption]
type OptionSpec[ModelType: GenericBaseModel] = ExecutableOption | OptionFactory[ModelType]


class SupportsWhere(Protocol):
    """Subset of SQLAlchemy statements that expose ``where``."""

    def where(self, *criteria: Any) -> Self: ...


class SupportsOptions(Protocol):
    """SQLAlchemy statements that accept loader options."""

    def options(self, *options: ExecutableOption) -> Self: ...



from typing import Any


def validate_exclusive_presence(first: Any | None, second: Any | None) -> None:
    """Validates that exactly one of `first` or `second` is provided.

    Raises a ValueError if none or both are provided.

    Args:
        first: The first value to check.
        second: The second value to check.

    """
    is_first = first is not None
    is_second = second is not None
    if is_first == is_second:  # Both are True or both are False
        msg = 'Exactly one of arguments must be provided, but not both.'
        raise ValueError(msg)



import typing
from collections.abc import Sequence
from datetime import datetime
from math import ceil
from typing import Annotated, Any, Literal
from uuid import UUID

from pydantic import BaseModel, ConfigDict, PlainSerializer

from notora.types import AnyIPAddress
from notora.v1.enums.base import OrderByDirections


def datetime_encoder(dec_value: datetime) -> float:
    return dec_value.timestamp()


timestamp = Annotated[datetime, PlainSerializer(datetime_encoder, return_type=float)]


class BaseResponseSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)


class BaseRequestSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)


class CreateUpdateMeta(BaseModel):
    created_at: datetime
    updated_at: datetime


class AdminMeta(CreateUpdateMeta):
    deleted_at: datetime | None = None


class AdminResponseSchema(BaseResponseSchema):
    id: UUID
    full_name: str | None = None
    unovay_name: str | None = None


class PersonalizedAdminMeta(CreateUpdateMeta):
    updated_by_user: AdminResponseSchema | None = None


class SetUpdatedBySchema(BaseRequestSchema):
    updated_at: datetime
    updated_by: UUID


class PaginationMetaSchema(BaseModel):
    limit: int
    total: int
    current_page: int
    last_page: int

    @classmethod
    def calculate(cls, total: int, limit: int, offset: int) -> 'PaginationMetaSchema':
        current_page = (offset // limit) + 1 if total >= limit else 1
        last_page = ceil(total / limit) if total > 0 else 1
        return cls(
            limit=limit,
            total=total,
            current_page=current_page,
            last_page=last_page,
        )


class PaginatedResponseSchema[T](BaseResponseSchema):
    meta: PaginationMetaSchema
    data: Sequence[T]


class ClientMeta(BaseRequestSchema):
    ip_address: Annotated[AnyIPAddress, PlainSerializer(str, return_type=str)] | None = None
    user_agent: str | None = None


class OrderBy(BaseModel):
    field: str
    direction: OrderByDirections = OrderByDirections.ASC
    model: type[Any] | None = None


FilterOp = Literal[
    'eq', '=', 'ilike', '~=', 'is', 'is_not', 'in', 'gt', '>', 'ge', '>=', 'lt', '<', 'le', '<='
]
filter_op_values = typing.get_args(FilterOp)


class Filter(BaseModel):
    field: str
    op: FilterOp = '='
    value: Any | None
    model: type[Any] | None = None


class OrFilterGroup(BaseModel):
    filters: list[Filter]


class BaseTokenSchema(BaseResponseSchema):
    sub: UUID
    iss: str
    nbf: timestamp
    exp: datetime
    iat: datetime

    @property
    def id(self) -> UUID:
        return self.sub


class BaseTokenParamsSchema(BaseRequestSchema): ...



import re
import uuid
from datetime import datetime
from typing import Any, ClassVar

from sqlalchemy import DateTime, ForeignKey, MetaData, Uuid, func
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    declared_attr,
    mapped_column,
    relationship,
)

POSTGRES_NAMING_CONVENTION = {
    'ix': '%(column_0_label)s_idx',
    'uq': '%(table_name)s_%(column_0_name)s_key',
    'ck': '%(table_name)s_%(constraint_name)s_check',
    'fk': '%(table_name)s_%(column_0_name)s_fkey',
    'pk': '%(table_name)s_pkey',
}
metadata = MetaData(naming_convention=POSTGRES_NAMING_CONVENTION)


class Base(DeclarativeBase):
    metadata = metadata


class CreatableMixin:
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=False),
        server_default=func.now(),
    )


class UpdatableMixin:
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=False),
        server_default=func.now(),
        onupdate=func.now(),
    )


class UpdatedByMixin:
    updated_by_fk_target: ClassVar[str | None] = None
    updated_by_fk_kwargs: ClassVar[dict[str, Any]] = {}

    @declared_attr
    @classmethod
    def updated_by(cls) -> Mapped[uuid.UUID | None]:
        if cls.updated_by_fk_target:
            return mapped_column(
                Uuid,
                ForeignKey(cls.updated_by_fk_target, **cls.updated_by_fk_kwargs),
                nullable=True,
            )
        return mapped_column(Uuid, nullable=True)


class UpdatedByUserMixin(UpdatedByMixin):
    updated_by_fk_target: ClassVar[str | None] = 'user.id'
    updated_by_fk_kwargs: ClassVar[dict[str, Any]] = {'use_alter': True}

    @declared_attr
    @classmethod
    def updated_by_user(cls) -> Mapped[Any | None]:
        return relationship('User', foreign_keys=[cls.updated_by])


class SoftDeletableMixin:
    deleted_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=False),
        nullable=True,
        default=None,
    )


class GenericBaseModel(Base, CreatableMixin):
    __abstract__ = True
    pk_type: ClassVar[Any] = Uuid
    pk_kwargs: ClassVar[dict[str, Any]] = {
        'server_default': func.gen_random_uuid(),
    }

    @declared_attr  # type: ignore[arg-type]
    def __tablename__(cls) -> str:  # noqa: N805
        return re.compile(r'(?<!^)(?=[A-Z])').sub('_', cls.__name__).lower()

    @declared_attr
    @classmethod
    def id(cls) -> Mapped[Any]:
        return mapped_column(cls.pk_type, primary_key=True, **cls.pk_kwargs)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} id={self.id}>'

    def to_dict(self) -> dict[str, Any]:
        return {field.name: getattr(self, field.name) for field in self.__table__.columns}


class BaseModel(GenericBaseModel, UpdatableMixin, SoftDeletableMixin):
    __abstract__ = True


class UpdatableModel(GenericBaseModel, UpdatableMixin):
    __abstract__ = True


class UpdatedByUserModel(GenericBaseModel, UpdatableMixin, UpdatedByUserMixin):
    __abstract__ = True


class SoftDeletableModel(GenericBaseModel, SoftDeletableMixin):
    __abstract__ = True


class AuditedBaseModel(BaseModel, UpdatedByUserMixin):
    __abstract__ = True



import logging
import re
from collections.abc import Iterable, Sequence
from typing import Any, ClassVar, overload
from uuid import UUID

from pydantic import BaseModel as BaseSchema
from sqlalchemy import Executable, ScalarResult, ScalarSelect, Select, exc
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.selectable import TypedReturnsRows

from notora.utils.time import now_without_tz
from notora.utils.validation import validate_exclusive_presence
from notora.v1.exceptions.common import AlreadyExistsError, FKNotFoundError, NotFoundError
from notora.v1.models.base import BaseModel, GenericBaseModel
from notora.v1.persistence.repos.base import BaseRepo, SoftDeletableRepo
from notora.v1.schemas.base import (
    BaseResponseSchema,
    Filter,
    OrderBy,
    OrFilterGroup,
    PaginatedResponseSchema,
    PaginationMetaSchema,
    SetUpdatedBySchema,
)

log = logging.getLogger(__name__)

type Filters = Filter | OrFilterGroup


class BaseService[  # noqa: PLR0904
    PKType,
    ModelClass: GenericBaseModel,
    ModelResponseSchema: BaseResponseSchema,
]:
    response_schema: type[ModelResponseSchema]
    repo: BaseRepo[PKType, ModelClass]

    _unique_violation_errors: ClassVar[dict[str, str]] = {}
    _fk_violation_errors: ClassVar[dict[str, str]] = {}

    _unique_constraint_pattern = re.compile(
        r'.*duplicate key value violates unique constraint "(?P<name>\w+)"',
    )
    _fk_constraint_pattern = re.compile(
        r'.*insert or update on table "(?P<table_name>\w+)" '
        r'violates foreign key constraint "(?P<fk_name>\w+)"',
    )

    @property
    def _not_found_error(self) -> str:
        return f'{self.__class__.__name__.removesuffix("Service")} not found.'

    def __init__(self, repo: BaseRepo[PKType, ModelClass]) -> None:
        self.repo = repo

    def serialize(self, db_result: ModelClass) -> ModelResponseSchema:
        return self.response_schema.model_validate(db_result)

    def serialize_many(self, db_result: Iterable[ModelClass]) -> list[ModelResponseSchema]:
        return [self.serialize(item) for item in db_result]

    async def list_raw(
        self,
        session: AsyncSession,
        filters: Iterable[Filters] = (),
        limit: int = 20,
        offset: int = 0,
        order_by: Iterable[OrderBy] = (),
        base_query: Select[tuple[Any]] | None = None,
    ) -> Iterable[ModelClass]:
        query = self.repo.list_(
            filters=filters,
            limit=limit,
            offset=offset,
            order_by=order_by,
            base_query=base_query,
        )
        return await session.scalars(query)

    async def list_(
        self,
        session: AsyncSession,
        filters: Iterable[Filters] = (),
        limit: int = 20,
        offset: int = 0,
        order_by: Iterable[OrderBy] = (),
        base_query: Select[tuple[Any]] | None = None,
    ) -> list[ModelResponseSchema]:
        results = await self.list_raw(
            session=session,
            filters=filters,
            limit=limit,
            offset=offset,
            order_by=order_by,
            base_query=base_query,
        )
        return self.serialize_many(results)

    async def bulk_create(
        self,
        db_session: AsyncSession,
        data: Sequence[BaseSchema],
    ) -> list[ModelResponseSchema]:
        query = self.repo.bulk_create([item.model_dump() for item in data])
        try:
            results = await db_session.scalars(query)
        except exc.IntegrityError as ex:
            raise self.__reraise_integrity_error(ex) from ex
        return self.serialize_many(results)

    @overload
    async def upsert(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        *,
        update_include_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
    ) -> ModelResponseSchema: ...

    @overload
    async def upsert(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        *,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
    ) -> ModelResponseSchema: ...

    async def upsert(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        *,
        update_include_fields: Iterable[str] | None = None,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
    ) -> ModelResponseSchema:
        validate_exclusive_presence(update_include_fields, update_exclude_fields)
        if update_include_fields is not None:
            result = await self.upsert_raw(
                db_session=db_session,
                data=data,
                update_include_fields=update_include_fields,
                index_elements=index_elements,
            )
        elif update_exclude_fields is not None:
            result = await self.upsert_raw(
                db_session=db_session,
                data=data,
                update_exclude_fields=update_exclude_fields,
                index_elements=index_elements,
            )
        return self.serialize(result)  # pyright: ignore [reportArgumentType], doesn't do conditional narrowing

    @overload
    async def upsert_raw(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        *,
        update_include_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
    ) -> ModelClass: ...

    @overload
    async def upsert_raw(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        *,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
    ) -> ModelClass: ...

    async def upsert_raw(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        *,
        update_include_fields: Iterable[str] | None = None,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
    ) -> ModelClass:
        validate_exclusive_presence(update_include_fields, update_exclude_fields)
        if update_include_fields is not None:
            query = self.repo.upsert(
                data=data.model_dump(),
                update_include_fields=update_include_fields,
                index_elements=index_elements,
            )
        elif update_exclude_fields is not None:
            query = self.repo.upsert(
                data=data.model_dump(),
                update_exclude_fields=update_exclude_fields,
                index_elements=index_elements,
            )
        result = await self.execute_for_one(db_session, query)
        return result

    async def paginate(
        self,
        session: AsyncSession,
        filters: Iterable[Filters] = (),
        limit: int = 20,
        offset: int = 0,
        order_by: Iterable[OrderBy] = (),
    ) -> PaginatedResponseSchema[ModelResponseSchema]:
        data = await self.list_(
            session=session,
            filters=filters,
            limit=limit,
            offset=offset,
            order_by=order_by,
        )
        meta = await self.get_filters_pagination_info(
            session=session,
            filters=filters,
            limit=limit,
            offset=offset,
        )
        return PaginatedResponseSchema(meta=meta, data=data)

    async def get_filters_pagination_info(
        self,
        session: AsyncSession,
        filters: Iterable[Filters] = (),
        limit: int = 20,
        offset: int = 0,
    ) -> PaginationMetaSchema:
        query = self.repo.number_of_records(filters)
        total = (await session.execute(query)).scalar_one()
        return PaginationMetaSchema.calculate(total, limit, offset)

    async def build_pagination_from_queries(
        self,
        db_session: AsyncSession,
        data_query: Executable,
        count_query: Executable,
        limit: int,
        offset: int,
    ) -> PaginatedResponseSchema[ModelResponseSchema]:
        data = self.serialize_many(await db_session.scalars(data_query))
        count = (await db_session.execute(count_query)).scalar_one()
        meta = PaginationMetaSchema.calculate(
            total=count,
            limit=limit,
            offset=offset,
        )
        return PaginatedResponseSchema(meta=meta, data=data)

    async def create(
        self,
        session: AsyncSession,
        data: BaseSchema,
    ) -> ModelResponseSchema:
        result = await self.create_raw(session, data)
        return self.serialize(result)

    async def create_raw(
        self,
        session: AsyncSession,
        data: BaseSchema,
    ) -> ModelClass:
        query = self.repo.create(data.model_dump())
        result = await self.execute_for_one(session, query)
        return result

    async def create_raw_or_skip(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        index_elements: Iterable[InstrumentedAttribute[Any]],
    ) -> ModelClass | None:
        query = self.repo.create_or_skip(data.model_dump(), index_elements)
        result = await self.execute(db_session, query)
        return result

    async def retrieve(self, session: AsyncSession, entity_id: PKType) -> ModelResponseSchema:
        entity = await self.retrieve_raw(session, entity_id)
        return self.serialize(entity)

    async def retrieve_raw(self, session: AsyncSession, entity_id: PKType) -> ModelClass:
        query = self.repo.retrieve(entity_id)
        result = await self.execute_for_one(session, query)
        return result

    async def retrieve_one_by(
        self,
        session: AsyncSession,
        filters: Iterable[Filters],
    ) -> ModelResponseSchema:
        data = await self.retrieve_one_raw_by(session, filters)
        return self.serialize(data)

    async def retrieve_one_raw_by(
        self,
        session: AsyncSession,
        filters: Iterable[Filters],
    ) -> ModelClass:
        query = self.repo.retrieve_by(filters)
        result = await self.execute_for_one(session, query)
        return result

    async def retrieve_all_by(
        self,
        db_session: AsyncSession,
        filters: Iterable[Filters] = (),
        order_by: Iterable[OrderBy] = (),
    ) -> list[ModelResponseSchema]:
        results = await self.retrieve_all_raw_by(db_session, filters, order_by)
        return self.serialize_many(results)

    async def retrieve_all_raw_by(
        self,
        db_session: AsyncSession,
        filters: Iterable[Filters] = (),
        order_by: Iterable[OrderBy] = (),
    ) -> ScalarResult[ModelClass]:
        query = self.repo.retrieve_by(filters, order_by)
        result = await db_session.scalars(query)
        return result

    async def hard_delete(self, session: AsyncSession, entity_id: PKType) -> None:
        query = self.repo.hard_delete(entity_id)
        await session.execute(query)

    async def update(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        entity_id: PKType | ScalarSelect[PKType],
    ) -> ModelResponseSchema:
        entity = await self.update_raw(db_session, data, entity_id)
        return self.serialize(entity)

    async def update_raw(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        entity_id: PKType | ScalarSelect[PKType],
    ) -> ModelClass:
        query = self.repo.update(entity_id, data.model_dump(exclude_unset=True))
        result = await self.execute_for_one(db_session, query)
        return result

    async def update_one_raw_by(
        self,
        db_session: AsyncSession,
        data: BaseSchema,
        filters: Iterable[Filters],
    ) -> ModelClass:
        query = self.repo.update_by(data.model_dump(exclude_unset=True), filters)
        result = await self.execute_for_one(db_session, query)
        return result

    async def set_updated_by(
        self,
        db_session: AsyncSession,
        updated_by: UUID,
        entity_id: PKType | ScalarSelect[PKType],
    ) -> ModelResponseSchema:
        return await self.update(
            db_session,
            SetUpdatedBySchema(updated_at=now_without_tz(), updated_by=updated_by),
            entity_id,
        )

    async def execute_for_one[T](
        self,
        db_session: AsyncSession,
        query: TypedReturnsRows[tuple[T]],
    ) -> T:
        result = await self.execute(db_session, query)
        if result is None:
            raise NotFoundError[PKType](self._not_found_error)
        return result

    async def execute[T](
        self,
        db_session: AsyncSession,
        query: TypedReturnsRows[tuple[T]],
    ) -> T | None:
        try:
            entity = await db_session.execute(query)
        except exc.IntegrityError as ex:
            raise self.__reraise_integrity_error(ex) from ex
        result = entity.unique().scalar_one_or_none()
        return result

    def __reraise_integrity_error(self, ex: exc.IntegrityError) -> Exception:
        if result := self._fk_constraint_pattern.match(ex.args[0]):
            fk_name = result.group('fk_name')
            table_name = result.group('table_name')
            msg = 'Related object not found.'
            return FKNotFoundError(
                self._fk_violation_errors.get(fk_name, msg),
                fk_name=fk_name,
                table_name=table_name,
            )
        if result := self._unique_constraint_pattern.match(ex.args[0]):
            constraint = result.group('name')
            msg = 'Entity already exists.'
            return AlreadyExistsError(
                self._unique_violation_errors.get(constraint, msg),
                constraint_name=constraint,
            )
        return ex


class SoftDeletableService[
    PKType,
    ModelClass: BaseModel,
    ModelResponseSchema: BaseResponseSchema,
](BaseService[PKType, ModelClass, ModelResponseSchema]):
    def __init__(self, repo: SoftDeletableRepo[PKType, ModelClass]) -> None:
        super().__init__(repo)
        self.repo: SoftDeletableRepo[PKType, ModelClass] = repo

    async def soft_delete(self, db_session: AsyncSession, entity_id: PKType) -> ModelClass:
        query = self.repo.soft_delete(entity_id)
        result = await self.execute_for_one(db_session, query)
        return result

    async def soft_delete_one_by(
        self,
        db_session: AsyncSession,
        filters: Iterable[Filters],
    ) -> ModelClass:
        query = self.repo.soft_delete_by(filters)
        result = await self.execute_for_one(db_session, query)
        return result

    async def soft_delete_all_raw_by(
        self,
        db_session: AsyncSession,
        filters: Iterable[Filters],
    ) -> Iterable[ModelClass]:
        query = self.repo.soft_delete_by(filters)
        result = await db_session.scalars(query)
        return result



from sqlalchemy.orm import Mapped

from notora.v1.models.base import BaseModel
from tests.v1.factories.base import SQLAlchemyBaseModelFactory


class User(BaseModel): ...


class MockModel(BaseModel):
    name: Mapped[str]


class MockModelFactory(SQLAlchemyBaseModelFactory[MockModel]): ...



from collections.abc import Iterable
from typing import Any, overload

from pydantic import BaseModel as PydanticModel
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.m2m import ManyToManySyncMixin
from notora.v2.services.mixins.payload import PayloadMixin
from notora.v2.services.mixins.serializer import SerializerProtocol
from notora.v2.services.mixins.updated_by import UpdatedByServiceMixin

__all__ = ['UpdateByFilterServiceMixin', 'UpdateServiceMixin']

class UpdateServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    ManyToManySyncMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    __type_params__: tuple[object, ...]

    async def update_raw(
        self,
        session: AsyncSession,
        pk: PKType,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType: ...
    @overload
    async def update(
        self,
        session: AsyncSession,
        pk: PKType,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: None = ...,
    ) -> DetailSchema: ...
    @overload
    async def update[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        pk: PKType,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[SchemaT],
    ) -> SchemaT: ...

class UpdateByFilterServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    __type_params__: tuple[object, ...]

    async def update_by_raw(
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType: ...
    @overload
    async def update_by(
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: None = ...,
    ) -> DetailSchema: ...
    @overload
    async def update_by[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[SchemaT],
    ) -> SchemaT: ...



from collections.abc import Iterable, Sequence
from typing import (
    Any,
    Protocol,
    cast,
    overload,
    override,
)

from sqlalchemy import (
    ScalarSelect,
    Select,
    UnaryExpression,
    and_,
    asc,
    delete,
    desc,
    false,
    func,
    null,
    or_,
    select,
    update,
)
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.base import ExecutableOption
from sqlalchemy.sql.dml import ReturningInsert
from sqlalchemy.sql.functions import now
from sqlalchemy.sql.selectable import TypedReturnsRows

from notora.utils.validation import validate_exclusive_presence
from notora.v1.enums.base import OrderByDirections
from notora.v1.models.base import BaseModel, GenericBaseModel
from notora.v1.schemas.base import Filter, OrderBy, OrFilterGroup

type Filters = Filter | OrFilterGroup


class HasWhere(Protocol):
    def where(self, *args: Any) -> 'HasWhere': ...


class BaseRepo[PKType, ModelType: GenericBaseModel]:
    load_options: Iterable[ExecutableOption] = []
    default_filters: Iterable[Filters] = []

    def __init__(self, model: type[ModelType], default_limit: int = 20):
        self.model = model
        self.default_limit = default_limit

    def select(self) -> Select[tuple[ModelType]]:
        return select(self.model).options(*self.load_options)

    def create_or_skip(
        self,
        data: dict[str, Any],
        index_elements: Iterable[InstrumentedAttribute[Any]],
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        where = None if not index_where else and_(*self._get_query_predicates(index_where))
        return (
            insert(self.model)
            .values(**data)
            .on_conflict_do_nothing(index_elements=index_elements, index_where=where)
            .returning(self.model)
            .options(*self.load_options)
        )

    def create(self, data: dict[str, Any]) -> ReturningInsert[tuple[ModelType]]:
        return insert(self.model).values(**data).returning(self.model).options(*self.load_options)

    def bulk_create(self, data: Sequence[dict[str, Any]]) -> ReturningInsert[tuple[ModelType]]:
        return insert(self.model).values(data).returning(self.model).options(*self.load_options)

    @overload
    def upsert(
        self,
        data: dict[str, Any],
        *,
        update_include_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]: ...

    @overload
    def upsert(
        self,
        data: dict[str, Any],
        *,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]: ...

    def upsert(
        self,
        data: dict[str, Any],
        *,
        update_include_fields: Iterable[str] | None = None,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        validate_exclusive_presence(update_include_fields, update_exclude_fields)
        if update_include_fields is not None:
            update_data = {
                key: value for key, value in data.items() if key in update_include_fields
            }
        elif update_exclude_fields is not None:
            update_data = {
                key: value for key, value in data.items() if key not in update_exclude_fields
            }

        where = None if not index_where else and_(*self._get_query_predicates(index_where))
        return (
            insert(self.model)
            .values(**data)
            .on_conflict_do_update(
                index_elements=index_elements or (self.model.id,),
                index_where=where,
                set_=update_data,
            )
            .returning(self.model)
            .options(*self.load_options)
        )

    def list_(
        self,
        base_query: Select[tuple[ModelType]] | None = None,
        filters: Iterable[Filters] = (),
        limit: int | None = None,
        offset: int = 0,
        order_by: Iterable[OrderBy] = (),
    ) -> Select[tuple[ModelType]]:
        query = base_query if base_query is not None else self.select()
        query = self.add_filters(query, filters)
        query = self.add_order_by(query, order_by)
        return query.limit(limit or self.default_limit).offset(offset)

    def retrieve(self, entity_id: PKType | ScalarSelect[PKType]) -> Select[tuple[ModelType]]:
        return self.list_(limit=1, filters=[Filter(field='id', op='=', value=entity_id)])

    def retrieve_by(
        self,
        filters: Iterable[Filters] = (),
        order_by: Iterable[OrderBy] = (),
    ) -> Select[tuple[ModelType]]:
        query = self.select()
        query = self.add_filters(query, filters)
        query = self.add_order_by(query, order_by)
        return query

    def update(
        self,
        entity_id: PKType | ScalarSelect[PKType],
        data: dict[str, object],
    ) -> TypedReturnsRows[tuple[ModelType]]:
        return self.update_by(
            data,
            filters=[Filter(field='id', op='eq', value=entity_id)],
        )

    def update_by(
        self,
        data: dict[str, object],
        filters: Iterable[Filters] = (),
    ) -> TypedReturnsRows[tuple[ModelType]]:
        query = update(self.model).values(**data)
        query = self.add_filters(query, filters)
        return query.returning(self.model).options(*self.load_options)

    def hard_delete(self, entity_id: PKType) -> TypedReturnsRows[tuple[ModelType]]:
        return self.hard_delete_by(
            filters=[Filter(field='id', op='eq', value=entity_id)],
        )

    def hard_delete_by(
        self, filters: Iterable[Filters] = ()
    ) -> TypedReturnsRows[tuple[ModelType]]:
        query = delete(self.model)
        query = self.add_filters(query, filters)
        return query.returning(self.model).options(*self.load_options)

    def number_of_records(self, filters: Iterable[Filters] = ()) -> Select[tuple[int]]:
        query = select(func.count()).select_from(self.model)
        query = self.add_filters(query, filters)
        return query

    def add_filters[T: HasWhere](self, query: T, filters: Iterable[Filters] = ()) -> T:
        filters = (*filters, *self.default_filters)
        for predicate in self._get_query_predicates(filters):
            query = cast(T, query.where(predicate))
        return query

    def add_order_by[T: Any](self, query: Select[T], order_by: Iterable[OrderBy]) -> Select[T]:
        orders: list[UnaryExpression[Any]] = []
        secondary_sort_model: type[ModelType] = self.model

        for order in order_by:
            target_model = order.model or self.model
            field = getattr(target_model, order.field)
            match order.direction:
                case OrderByDirections.ASC:
                    direction_func = asc
                case OrderByDirections.DESC:
                    direction_func = desc
            orders.append(direction_func(field))
            if order.model is not None:
                secondary_sort_model = order.model

        return query.order_by(*orders, secondary_sort_model.id)

    def _get_query_predicates(self, filters: Iterable[Filters]) -> Iterable[Any]:  # noqa: C901, PLR0912
        predicates = []
        for filter_ in filters:
            match filter_:
                case OrFilterGroup(filters=or_filters):
                    or_predicates = or_(*self._get_query_predicates(or_filters))
                    predicates.append(or_predicates)
                case Filter():
                    target_model = filter_.model or self.model
                    field = getattr(target_model, filter_.field)
                    match filter_:
                        case Filter(op='eq' | '=', value=None):
                            predicates.append(field.is_(None))
                        case Filter(op='is', value=v1):
                            predicates.append(field.is_(v1))
                        case Filter(value=None):
                            continue
                        case Filter(op='eq' | '=', value=v2):
                            predicates.append(field == v2)
                        case Filter(op='ilike' | '~=', value=v3):
                            predicates.append(field.ilike(f'%{v3}%'))
                        case Filter(op='in', value=[] | () | set(())):
                            predicates.append(false())
                        case Filter(op='in', value=v4):
                            predicates.append(field.in_(v4))
                        case Filter(op='gt' | '>', value=v5):
                            predicates.append(field > v5)
                        case Filter(op='ge' | '>=', value=v6):
                            predicates.append(field >= v6)
                        case Filter(op='lt' | '<', value=v7):
                            predicates.append(field < v7)
                        case Filter(op='le' | '<=', value=v8):
                            predicates.append(field <= v8)
        return predicates


class SoftDeletableRepo[
    PKType,
    ModelType: BaseModel,
](BaseRepo[PKType, ModelType]):
    default_filters = (Filter(field='deleted_at', op='is', value=null()),)

    @override
    def select(self) -> Select[tuple[ModelType]]:
        return self.add_filters(super().select())

    @override
    def create_or_skip(
        self,
        data: dict[str, Any],
        index_elements: Iterable[InstrumentedAttribute[Any]],
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        index_where = index_where or (Filter(field='deleted_at', op='is', value=null()),)
        return super().create_or_skip(
            data=data, index_elements=index_elements, index_where=index_where
        )

    @overload
    def upsert(
        self,
        data: dict[str, Any],
        *,
        update_include_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]: ...

    @overload
    def upsert(
        self,
        data: dict[str, Any],
        *,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]: ...

    @override
    def upsert(
        self,
        data: dict[str, Any],
        *,
        update_include_fields: Iterable[str] | None = None,
        update_exclude_fields: Iterable[str] | None = None,
        index_elements: Iterable[InstrumentedAttribute[Any]] | None = None,
        index_where: Iterable[Filters] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        index_where = index_where or (Filter(field='deleted_at', op='is', value=null()),)
        validate_exclusive_presence(update_include_fields, update_exclude_fields)
        if update_include_fields is not None:
            query = super().upsert(
                data=data,
                update_include_fields=update_include_fields,
                index_elements=index_elements,
                index_where=index_where,
            )
        elif update_exclude_fields is not None:
            query = super().upsert(
                data=data,
                update_exclude_fields=update_exclude_fields,
                index_elements=index_elements,
                index_where=index_where,
            )
        return query

    def soft_delete(
        self,
        entity_id: PKType,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        return self.soft_delete_by([Filter(field='id', op='eq', value=entity_id)])

    def soft_delete_by(
        self,
        filters: Iterable[Filters] = (),
    ) -> TypedReturnsRows[tuple[ModelType]]:
        query = update(self.model).values({'deleted_at': now()})
        query = self.add_filters(query, filters)
        return query.returning(self.model)



from collections.abc import Iterable

from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import SoftDeleteRepositoryProtocol
from notora.v2.repositories.types import FilterSpec
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin


class DeleteServiceMixin[PKType, ModelType: GenericBaseModel](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
):
    async def delete(self, session: AsyncSession, pk: PKType) -> None:
        await self.execute(session, self.repo.delete(pk))

    async def delete_by(
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
    ) -> None:
        await self.execute(session, self.repo.delete_by(filters=filters))


class SoftDeleteServiceMixin[PKType, ModelType: GenericBaseModel](
    DeleteServiceMixin[PKType, ModelType],
):
    repo: SoftDeleteRepositoryProtocol[PKType, ModelType]

    async def soft_delete(self, session: AsyncSession, pk: PKType) -> None:
        await self.execute(session, self.repo.soft_delete(pk))

    async def soft_delete_by(
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
    ) -> None:
        await self.execute(session, self.repo.soft_delete_by(filters=filters))



from collections.abc import Sequence
from dataclasses import dataclass

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec, OrderSpec


@dataclass(slots=True)
class RepoConfig[ModelType: GenericBaseModel]:
    default_limit: int | None = None
    default_options: Sequence[OptionSpec[ModelType]] | None = None
    default_filters: Sequence[FilterSpec[ModelType]] | None = None
    default_ordering: Sequence[OrderSpec[ModelType]] | None = None
    fallback_sort_attribute: str | None = None
    pk_attribute: str | None = None
    apply_soft_delete_filter: bool | None = None



from collections.abc import Iterable
from dataclasses import dataclass

from sqlalchemy import Select

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import (
    DEFAULT_LIMIT,
    DefaultLimit,
    FilterSpec,
    OptionSpec,
    OrderSpec,
)


@dataclass(slots=True)
class QueryParams[ModelType: GenericBaseModel]:
    filters: Iterable[FilterSpec[ModelType]] | None = None
    ordering: Iterable[OrderSpec[ModelType]] | None = None
    options: Iterable[OptionSpec[ModelType]] | None = None
    limit: int | DefaultLimit | None = DEFAULT_LIMIT
    offset: int = 0
    base_query: Select[tuple[ModelType]] | None = None


@dataclass(slots=True)
class PaginationParams[ModelType: GenericBaseModel]:
    filters: Iterable[FilterSpec[ModelType]] | None = None
    ordering: Iterable[OrderSpec[ModelType]] | None = None
    options: Iterable[OptionSpec[ModelType]] | None = None
    limit: int = 20
    offset: int = 0
    base_query: Select[tuple[ModelType]] | None = None



from uuid import UUID

from notora.v1.services.base import SoftDeletableService
from tests.v1.test_integration.mocks.model import MockModel
from tests.v1.test_integration.mocks.schema import MockModelResponseSchema


class MockService(SoftDeletableService[UUID, MockModel, MockModelResponseSchema]): ...



from collections.abc import Iterable, Sequence
from typing import Any, Protocol, cast

from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.selectable import TypedReturnsRows

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.config import RepoConfig
from notora.v2.repositories.mixins import (
    CountableMixin,
    CreateMixin,
    CreateOrSkipMixin,
    DeleteMixin,
    RetrievableMixin,
    SoftDeleteMixin,
    UpdateMixin,
    UpsertMixin,
)
from notora.v2.repositories.params import QueryParams
from notora.v2.repositories.types import DefaultLimit, FilterSpec, OptionSpec, OrderSpec


class Repository[PKType, ModelType: GenericBaseModel](
    RetrievableMixin[PKType, ModelType],
    CreateMixin[ModelType],
    UpsertMixin[PKType, ModelType],
    CreateOrSkipMixin[ModelType],
    UpdateMixin[PKType, ModelType],
    DeleteMixin[PKType, ModelType],
    CountableMixin[ModelType],
):
    """Composition-friendly base repository built out of mixins."""

    def __init__(  # noqa: C901
        self, model: type[ModelType], *, config: RepoConfig[ModelType] | None = None
    ) -> None:
        self.model = model
        if config is None:
            return
        if config.default_limit is not None:
            self.default_limit = config.default_limit
        if config.default_options is not None:
            self.default_options = config.default_options
        if config.default_filters is not None:
            self.default_filters = config.default_filters
        if config.default_ordering is not None:
            self.default_ordering = config.default_ordering
        if config.fallback_sort_attribute is not None:
            self.fallback_sort_attribute = config.fallback_sort_attribute
        if config.pk_attribute is not None:
            self.pk_attribute = config.pk_attribute


class SoftDeleteRepository[PKType, ModelType: GenericBaseModel](
    RetrievableMixin[PKType, ModelType],
    CreateMixin[ModelType],
    UpsertMixin[PKType, ModelType],
    CreateOrSkipMixin[ModelType],
    SoftDeleteMixin[PKType, ModelType],
    DeleteMixin[PKType, ModelType],
    CountableMixin[ModelType],
):
    """Repository variant with soft-delete helpers."""

    apply_soft_delete_filter: bool = True

    def _soft_delete_filter(self) -> FilterSpec[ModelType]:
        column = cast(InstrumentedAttribute[Any], getattr(self.model, self.deleted_attribute))
        return column.is_(None)

    def __init__(  # noqa: C901
        self, model: type[ModelType], *, config: RepoConfig[ModelType] | None = None
    ) -> None:
        self.model = model
        apply_soft_delete_filter = self.apply_soft_delete_filter
        if config is not None:
            if config.default_limit is not None:
                self.default_limit = config.default_limit
            if config.default_options is not None:
                self.default_options = config.default_options
            if config.default_filters is not None:
                self.default_filters = config.default_filters
            if config.default_ordering is not None:
                self.default_ordering = config.default_ordering
            if config.fallback_sort_attribute is not None:
                self.fallback_sort_attribute = config.fallback_sort_attribute
            if config.pk_attribute is not None:
                self.pk_attribute = config.pk_attribute
            if config.apply_soft_delete_filter is not None:
                apply_soft_delete_filter = config.apply_soft_delete_filter
        if apply_soft_delete_filter:
            base_filters = tuple(self.default_filters)
            self.default_filters = (*base_filters, self._soft_delete_filter())


class RepositoryProtocol[PKType, ModelType: GenericBaseModel](Protocol):
    model: type[ModelType]
    pk_attribute: str
    default_limit: int

    def list(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = ...,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def list_by_params(
        self, params: QueryParams[ModelType]
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def count(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[int]]: ...

    def retrieve(
        self,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def retrieve_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def retrieve_one_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def create(
        self,
        payload: dict[str, Any],
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def create_or_skip(
        self,
        payload: dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]],
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def upsert(
        self,
        payload: dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]] | None = None,
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        update_only: Sequence[str] | None = None,
        update_exclude: Sequence[str] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def update(
        self,
        pk: PKType,
        payload: dict[str, Any],
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def update_by(
        self,
        payload: dict[str, Any],
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def delete(
        self,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def delete_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...


class SoftDeleteRepositoryProtocol[PKType, ModelType: GenericBaseModel](
    RepositoryProtocol[PKType, ModelType],
    Protocol,
):
    def soft_delete(
        self,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...

    def soft_delete_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]: ...



from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import RepositoryProtocol, SoftDeleteRepositoryProtocol
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.config import ServiceConfig
from notora.v2.services.mixins.create import CreateOrSkipServiceMixin, CreateServiceMixin
from notora.v2.services.mixins.delete import DeleteServiceMixin, SoftDeleteServiceMixin
from notora.v2.services.mixins.pagination import PaginationServiceMixin
from notora.v2.services.mixins.retrieval import RetrievalServiceMixin
from notora.v2.services.mixins.serializer import SerializerMixin
from notora.v2.services.mixins.update import UpdateByFilterServiceMixin, UpdateServiceMixin
from notora.v2.services.mixins.upsert import UpsertServiceMixin


class RepositoryService[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SerializerMixin[ModelType, DetailSchema, ListSchema],
    PaginationServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    RetrievalServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    CreateServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    CreateOrSkipServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    UpsertServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    UpdateServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    UpdateByFilterServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
    DeleteServiceMixin[PKType, ModelType],
):
    """Turnkey async service that glues repository access and serialization together."""

    def __init__(
        self,
        repo: RepositoryProtocol[PKType, ModelType],
        *,
        config: ServiceConfig[DetailSchema, ListSchema] | None = None,
    ) -> None:
        self.repo = repo
        if config is None:
            return
        if config.detail_schema is not None:
            self.detail_schema = config.detail_schema
        if config.list_schema is not None:
            self.list_schema = config.list_schema


class SoftDeleteRepositoryService[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    RepositoryService[PKType, ModelType, DetailSchema, ListSchema],
    SoftDeleteServiceMixin[PKType, ModelType],
):
    """Repository service variant that exposes soft-delete helpers."""

    def __init__(
        self,
        repo: SoftDeleteRepositoryProtocol[PKType, ModelType],
        *,
        config: ServiceConfig[DetailSchema, ListSchema] | None = None,
    ) -> None:
        super().__init__(repo, config=config)


type RepositoryServiceD[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
] = RepositoryService[PKType, ModelType, DetailSchema, DetailSchema]


type SoftDeleteRepositoryServiceD[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
] = SoftDeleteRepositoryService[PKType, ModelType, DetailSchema, DetailSchema]



from uuid import UUID

from sqlalchemy import Boolean, ForeignKey, String, UniqueConstraint
from sqlalchemy.dialects.postgresql import UUID as PGUUID
from sqlalchemy.orm import Mapped, mapped_column

from notora.v2.models.base import (
    GenericBaseModel,
    SoftDeletableMixin,
    UpdatableMixin,
    UpdatedByMixin,
)


class V2User(GenericBaseModel, UpdatableMixin, UpdatedByMixin, SoftDeletableMixin):
    __table_args__ = (UniqueConstraint('email', name='v2_user_email_key'),)

    email: Mapped[str] = mapped_column(String, nullable=False)
    name: Mapped[str] = mapped_column(String, nullable=False)
    is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)


class V2Role(GenericBaseModel):
    name: Mapped[str] = mapped_column(String, nullable=False)


class V2UserRole(GenericBaseModel):
    __table_args__ = (UniqueConstraint('user_id', 'role_id', name='v2_user_role_unique'),)

    user_id: Mapped[UUID] = mapped_column(
        PGUUID,
        ForeignKey('v2_user.id', ondelete='CASCADE'),
        nullable=False,
    )
    role_id: Mapped[UUID] = mapped_column(
        PGUUID,
        ForeignKey('v2_role.id', ondelete='CASCADE'),
        nullable=False,
    )


class V2Profile(GenericBaseModel):
    user_id: Mapped[UUID] = mapped_column(
        PGUUID,
        ForeignKey('v2_user.id'),
        nullable=False,
    )
    bio: Mapped[str] = mapped_column(String, nullable=False)



from uuid import UUID

from notora.v2.services.base import RepositoryService, SoftDeleteRepositoryService
from notora.v2.services.mixins.m2m import ManyToManyRelation
from tests.v2.test_integration.mocks.model import V2Profile, V2User, V2UserRole
from tests.v2.test_integration.mocks.schema import (
    V2ProfileResponseSchema,
    V2UserResponseSchema,
)


class V2UserService(SoftDeleteRepositoryService[UUID, V2User, V2UserResponseSchema]):
    many_to_many_relations = (
        ManyToManyRelation[V2UserRole](
            payload_field='role_ids',
            association_model=V2UserRole,
            left_key=V2UserRole.user_id,
            right_key=V2UserRole.role_id,
        ),
    )


class V2ProfileService(RepositoryService[UUID, V2Profile, V2ProfileResponseSchema]): ...



from collections.abc import Iterable, Sequence
from typing import Any

from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.params import QueryParams
from notora.v2.repositories.types import (
    DEFAULT_LIMIT,
    DefaultLimit,
    FilterSpec,
    OptionSpec,
    OrderSpec,
)
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.serializer import SerializerProtocol

__all__ = ['ListResponse', 'ListingServiceMixin']

ListResponse = list


class ListingServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def list_raw(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = DEFAULT_LIMIT,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
    ) -> Sequence[ModelType]:
        query = self.repo.list(
            filters=filters,
            limit=limit,
            offset=offset,
            ordering=ordering,
            options=options,
            base_query=base_query,
        )
        return await self.execute_scalars_all(session, query)

    async def list(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = DEFAULT_LIMIT,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
        schema: type[ListSchema] | None = None,
    ) -> ListResponse[ListSchema]:
        rows = await self.list_raw(
            session,
            filters=filters,
            limit=limit,
            offset=offset,
            ordering=ordering,
            options=options,
            base_query=base_query,
        )
        return self.serialize_many(rows, schema=schema)

    async def list_raw_params(
        self,
        session: AsyncSession,
        params: QueryParams[ModelType],
    ) -> Sequence[ModelType]:
        return await self.list_raw(
            session,
            filters=params.filters,
            limit=params.limit,
            offset=params.offset,
            ordering=params.ordering,
            options=params.options,
            base_query=params.base_query,
        )

    async def list_params(
        self,
        session: AsyncSession,
        params: QueryParams[ModelType],
        *,
        schema: type[ListSchema] | None = None,
    ) -> ListResponse[ListSchema]:
        rows = await self.list_raw_params(session, params)
        return self.serialize_many(rows, schema=schema)



from collections.abc import Iterable
from typing import Protocol, cast

from notora.v2.models.base import GenericBaseModel
from notora.v2.schemas.base import BaseResponseSchema

__all__ = ['SerializerMixin', 'SerializerProtocol']


class SerializerProtocol[
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](Protocol):
    def serialize_one(
        self,
        obj: ModelType,
        *,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema: ...

    def serialize_many(
        self,
        objs: Iterable[ModelType],
        *,
        schema: type[ListSchema] | None = None,
        prefer_list_schema: bool = True,
    ) -> list[ListSchema]: ...


class SerializerMixin[
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
]:
    detail_schema: type[DetailSchema] | None = None
    list_schema: type[ListSchema] | None = None

    def serialize_one(
        self,
        obj: ModelType,
        *,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        if schema is None:
            schema = self.detail_schema
        if schema is None:
            msg = 'schema is required for serialized methods; use *_raw or set detail_schema.'
            raise ValueError(msg)
        return schema.model_validate(obj)

    def serialize_many(
        self,
        objs: Iterable[ModelType],
        *,
        schema: type[ListSchema] | None = None,
        prefer_list_schema: bool = True,
    ) -> list[ListSchema]:
        if schema is None and prefer_list_schema:
            schema = self.list_schema
            if schema is None:
                schema = cast(type[ListSchema] | None, self.detail_schema)
        if schema is None:
            msg = 'schema is required for serialized methods; use *_raw or set list_schema.'
            raise ValueError(msg)
        return [schema.model_validate(obj) for obj in objs]



from collections.abc import Iterable, Sequence
from typing import Any, overload

from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.params import QueryParams
from notora.v2.repositories.types import (
    DEFAULT_LIMIT,
    DefaultLimit,
    FilterSpec,
    OptionSpec,
    OrderSpec,
)
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.serializer import SerializerProtocol

__all__ = ['ListResponse', 'ListingServiceMixin']

type ListResponse[ListSchema: BaseResponseSchema] = list[ListSchema]

class ListingServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    __type_params__: tuple[object, ...]

    async def list_raw(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = DEFAULT_LIMIT,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
    ) -> Sequence[ModelType]: ...
    @overload
    async def list(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = DEFAULT_LIMIT,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
        schema: None = ...,
    ) -> ListResponse[ListSchema]: ...
    @overload
    async def list[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = DEFAULT_LIMIT,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
        schema: type[SchemaT],
    ) -> ListResponse[SchemaT]: ...
    async def list_raw_params(
        self,
        session: AsyncSession,
        params: QueryParams[ModelType],
    ) -> Sequence[ModelType]: ...
    @overload
    async def list_params(
        self,
        session: AsyncSession,
        params: QueryParams[ModelType],
        *,
        schema: None = ...,
    ) -> ListResponse[ListSchema]: ...
    @overload
    async def list_params[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        params: QueryParams[ModelType],
        *,
        schema: type[SchemaT],
    ) -> ListResponse[SchemaT]: ...
